@@ -83,17 +83,63 @@ knit2pdf = function(
   with_ext(out, 'pdf')
 }
 
+#' Convert an \file{Rnw} document to PDF
+#'
+#' Call \code{\link{knit}()} to compile the \file{.Rnw} input to \file{.tex},
+#' and then \code{tinytex::\link[tinytex]{latexmk}()} to convert \file{.tex} to
+#' \file{.pdf}.
+#'
+#' This function is similar to \code{\link{knit2pdf}()}, with the following differences:
+#' \enumerate{
+#'   \item The default compiler is "xelatex" instead of "pdflatex".
+#'   \item \code{output} uses the file extension ".pdf" instead of ".tex".
+#'   \item Before knitting, it tries to remove the \code{output} file and will throw a clear error if the file cannot be removed.
+#'   \item \code{output} could be under any dir, not necessarily the same directory as \code{input}.
+#'   \item It cleans up intermediate files by default, including the ".tex" file.
+#'   \item It stops knitting when any error occurs (by setting the chunk option \code{error = FALSE}).
+#' }
+#' @inheritParams knit
+#' @param output Path of the PDF output file. By default, it uses the same name
+#'   as the \code{input}, but changes the file extension to ".pdf".
+#' @param compiler,... The LaTeX engine and other arguments to be passed to
+#'   \code{tinytex::\link[tinytex]{latexmk}()}. The default compiler is
+#'   \code{xelatex}.
+#' @param clean If \code{TRUE}, the intermediate files will be removed.
+#' @param error If \code{FALSE}, knitting stops when any error occurs.
+#' @return The \code{output} file path.
+#' @export
+rnw2pdf = function(
+  input, output = with_ext(input, 'pdf'), compiler = 'xelatex',
+  envir = parent.frame(), quiet = FALSE, clean = TRUE, error = FALSE, ...
+) {
+  # On Windows, when tweaking the content, users may forget to close the PDF
+  # file (thus can't be written). Since knitting may take quite some time, it's
+  # better to check the write permission of the output file in advance.
+  file.remove(output)
+  if (xfun::file_exists(output)) stop(
+    "The file '", output, "' cannot be removed (may be locked by a PDF reader)."
+  )
+  old = opts_chunk$set(error = error)
+  on.exit(opts_chunk$set(old), add = TRUE)
+  file_tex = knit(input, envir = envir, quiet = quiet)
+  if (clean) on.exit(file.remove(file_tex), add = TRUE)
+  file_pdf = tinytex::latexmk(file_tex, engine = compiler, clean = clean, ...)
+  if (!xfun::same_path(output, file_pdf)) file.rename(file_pdf, output)
+  output
+}
+
 #' Convert markdown to HTML using knit() and markdownToHTML()
 #'
 #' This is a convenience function to knit the input markdown source and call
-#' \code{markdown::\link{markdownToHTML}()} in the \pkg{markdown} package to
-#' convert the result to HTML.
+#' \code{markdown::\link[markdown]{markdownToHTML}()} in the \pkg{markdown}
+#' package to convert the result to HTML.
 #' @inheritParams knit
-#' @param ... Options passed to \code{markdown::\link{markdownToHTML}()}.
+#' @param ... Options passed to
+#'   \code{markdown::\link[markdown]{markdownToHTML}()}.
 #' @param force_v1 Boolean; whether to force rendering the input document as an
 #'   R Markdown v1 document, even if it is for v2.
 #' @export
-#' @seealso \code{\link{knit}}, \code{markdown::\link{markdownToHTML}}
+#' @seealso \code{\link{knit}}, \code{markdown::\link[markdown]{markdownToHTML}}
 #' @return If the argument \code{text} is NULL, a character string (HTML code)
 #'   is returned; otherwise the result is written into a file and the filename
 #'   is returned.
@@ -111,9 +157,9 @@ knit2html = function(
   input, output = NULL, ..., envir = parent.frame(), text = NULL,
   quiet = FALSE, encoding = 'UTF-8', force_v1 = getOption('knitr.knit2html.force_v1', FALSE)
 ) {
-  # packages containing vignettes using R Markdown v1 should declare dependency
-  # on 'markdown' in DESCRIPTION (typically in Suggests)
-  test_vig_dep('markdown')
+  if (is_cran_check() && !has_package('markdown'))
+    return(vweave_empty(input, .reason = 'markdown'))
+
   if (!force_v1 && is.null(text)) {
     if (length(grep('^---\\s*$', head(read_utf8(input), 1)))) warning2(
       'It seems you should call rmarkdown::render() instead of knitr::knit2html() ',
